//////////////////////////////////////////////////////////////////////////////
//																			//
//				    	Atmel AT91SAM7X CAN Port Routines					//
//																			//
//////////////////////////////////////////////////////////////////////////////
//
//	Author:	Andy Kunz
//	Date:	2-20-08
//	Mods:	Adapt from UART.C
//
//			NOTE:  THIS FILE MUST BE COMPILED IN ARM MODE!
//

#define ATCAN_C

#include <string.h>

#include "defs.h"
#include "target.h"
#include "datatypes.h"
#include "at91sam7x256.h"

#include "pio.h"
#include "atcan.h"

#include "board.h"
#include "FreeRTOS.h"
#include "queue.h"
#include "task.h"

//////////////////////////////////////////////////////////////////////////////
//
//							Data Buffers
//
//////////////////////////////////////////////////////////////////////////////
//
//	Author:		Andy Kunz
//	Date:		2-20-08
//	Modifications:	none
//

#define	CAN_RX_QUEUE_SIZE		(16)					// Number of messages in received msg queue
#define	CAN_TX_QUEUE_SIZE		(8)						// Number of messages in outgoing msg queue
#define	RX_BLOCKTIME			(1)						// Blocking time on receive queue
#define	TX_BLOCKTIME			(1)						// Blocking time on transmit queue

static
xQueueHandle	rx_queue,								// Input buffer
				tx_queue;								// Output buffer

//////////////////////////////////////////////////////////////////////////////
//
//							Mailbox & CAN_BR configuration
//
//////////////////////////////////////////////////////////////////////////////
//
//		NOTE:	This section shows how you need to set up the ISR for
//				processing mailbox interrupts.
//
//////////////////////////////////////////////////////////////////////////////
//
//	Author:		Andy Kunz
//	Date:		2-20-08
//	Modifications:	none
//
// Mailbox 0:		Receive
// Mailbox 1:		Receive
// Mailbox 2:		Receive
// Mailbox 3:		Receive
// Mailbox 4:		Receive
// Mailbox 5:		Receive
// Mailbox 6:		Receive with Overwrite
// Mailbox 7:		Transmit

#define	USE_CYGNUS_PARAMS			(FALSE)

#if (USE_CYGNUS_PARAMS)
//	Taken from can_at91sam7.c (Cygnus)
//
//	#define CAN_BR_TBL_ENTRY(_brp_, _propag_, _phase1_, _phase2_, _sjw_) <backslash>
//		((_brp_ << 16) | (_propag_ << 8) | (_phase2_) | (_phase1_ << 4) | (_sjw_ << 12))
//
//    CAN_BR_TBL_ENTRY(0x17, 0x04, 0x07, 0x01, 0), // 125 kbaud

#define CAN_BR_BRP			(0x17)
#define	CAN_BR_PROPAG		(0x04)
#define	CAN_BR_PHASE_1		(0x07)
#define	CAN_BR_PHASE_2		(0x01)
#define	CAN_BR_SJW			(0)
#define	CAN_BR_SMP			(1)

#else
//	Taken from sample code on www.AT91.com forum by Gerd

#define	CAN_BR_125K			(0x00173255UL)

#endif

//////////////////////////////////////////////////////////////////////////////
//
//					Interrupt Service Routine
//
//////////////////////////////////////////////////////////////////////////////
//
//	Author:		Andy Kunz
//	Date:		4-1-08
//	Modifications:	Don't check queue before attempting to read it (Transmit mbox)
//
//	Author:		Andy Kunz
//	Date:		3-31-08
//	Modifications:	Read/write full 32-bit index in msg struct
//
//	Author:		Andy Kunz
//	Date:		3-26-08
//	Modifications:	When the tx queue is empty, disable the tx interrupt
//
//	Author:		Andy Kunz
//	Date:		2-20-08
//	Modifications:	none
//

static void atcan_isr (void);							// CAN ISR
static void atcan_isr (void)							// CAN ISR
	{
	UINT32	status,										// CAN module status
			mbstat;										// Mailbox Status
	INT32	awakened_by_tx,								// Tells why we woke up to the RTOS
			awakened_by_rx;
	ATCANMSG msg;										// Message holder

	awakened_by_tx =
	awakened_by_rx = FALSE;

	status = *AT91C_CAN_SR;								// Get current status
	status &= *AT91C_CAN_IMR;							// Clear unavailable interrupt flags

	//////////////////////////////////////////////////////
	//
	//				Receive Mailboxes
	//
	//////////////////////////////////////////////////////

	if (status & AT91C_CAN_MB0)							// Process Receive Mailbox 0 event
		{
		mbstat = *AT91C_CAN_MB0_MSR;
		if (mbstat & AT91C_CAN_MRDY)					// If the event was a received msg
			{											//	 Receive msg and stick into queue
			msg.msglen = (mbstat >> 16) & 0x0F;			//		Grab message length
			msg.index  = *AT91C_CAN_MB0_MID;			//		Grab message ID
			msg.mailbox= 0;								//		Remember which mailbox it came in on
			msg.payload_lo = *AT91C_CAN_MB0_MDL;		//		Grab message data
			msg.payload_hi = *AT91C_CAN_MB0_MDH;
			*AT91C_CAN_MB0_MCR = AT91C_CAN_MTCR;		//	Allow another message to be received here
			if (xQueueSendFromISR (rx_queue, &msg, FALSE))//	Stick message into queue for application
				{
				awakened_by_rx = TRUE;
				}
			}
		}

	if (status & AT91C_CAN_MB1)							// Process Receive Mailbox 1 event
		{
		mbstat = *AT91C_CAN_MB1_MSR;
		if (mbstat & AT91C_CAN_MRDY)
			{
			msg.msglen = (mbstat >> 16) & 0x0F;
			msg.index  = *AT91C_CAN_MB1_MID;
			msg.mailbox= 1;
			msg.payload_lo = *AT91C_CAN_MB1_MDL;
			msg.payload_hi = *AT91C_CAN_MB1_MDH;
			*AT91C_CAN_MB1_MCR = AT91C_CAN_MTCR;
			if (xQueueSendFromISR (rx_queue, &msg, FALSE))
				{
				awakened_by_rx = TRUE;
				}
			}
		}

	if (status & AT91C_CAN_MB2)							// Process Receive Mailbox 2 event
		{
		mbstat = *AT91C_CAN_MB2_MSR;
		if (mbstat & AT91C_CAN_MRDY)
			{
			msg.msglen = (mbstat >> 16) & 0x0F;
			msg.index  = *AT91C_CAN_MB1_MID;
			msg.mailbox= 2;
			msg.payload_lo = *AT91C_CAN_MB2_MDL;
			msg.payload_hi = *AT91C_CAN_MB2_MDH;
			*AT91C_CAN_MB2_MCR = AT91C_CAN_MTCR;
			if (xQueueSendFromISR (rx_queue, &msg, FALSE))
				{
				awakened_by_rx = TRUE;
				}
			}
		}

	if (status & AT91C_CAN_MB3)							// Process Receive Mailbox 3 event
		{
		mbstat = *AT91C_CAN_MB3_MSR;
		if (mbstat & AT91C_CAN_MRDY)
			{
			msg.msglen = (mbstat >> 16) & 0x0F;
			msg.index  = *AT91C_CAN_MB1_MID;
			msg.mailbox= 3;
			msg.payload_lo = *AT91C_CAN_MB3_MDL;
			msg.payload_hi = *AT91C_CAN_MB3_MDH;
			*AT91C_CAN_MB3_MCR = AT91C_CAN_MTCR;
			if (xQueueSendFromISR (rx_queue, &msg, FALSE))
				{
				awakened_by_rx = TRUE;
				}
			}
		}

	if (status & AT91C_CAN_MB4)							// Process Receive Mailbox 4 event
		{
		mbstat = *AT91C_CAN_MB4_MSR;
		if (mbstat & AT91C_CAN_MRDY)
			{
			msg.msglen = (mbstat >> 16) & 0x0F;
			msg.index  = *AT91C_CAN_MB1_MID;
			msg.mailbox= 4;
			msg.payload_lo = *AT91C_CAN_MB4_MDL;
			msg.payload_hi = *AT91C_CAN_MB4_MDH;
			*AT91C_CAN_MB4_MCR = AT91C_CAN_MTCR;
			if (xQueueSendFromISR (rx_queue, &msg, FALSE))
				{
				awakened_by_rx = TRUE;
				}
			}
		}

	if (status & AT91C_CAN_MB5)							// Process Receive Mailbox 5 event
		{
		mbstat = *AT91C_CAN_MB5_MSR;
		if (mbstat & AT91C_CAN_MRDY)
			{
			msg.msglen = (mbstat >> 16) & 0x0F;
			msg.index  = *AT91C_CAN_MB1_MID;
			msg.mailbox= 5;
			msg.payload_lo = *AT91C_CAN_MB5_MDL;
			msg.payload_hi = *AT91C_CAN_MB5_MDH;
			*AT91C_CAN_MB5_MCR = AT91C_CAN_MTCR;
			if (xQueueSendFromISR (rx_queue, &msg, FALSE))
				{
				awakened_by_rx = TRUE;
				}
			}
		}

	if (status & AT91C_CAN_MB6)							// Process Receive Mailbox 6 event
		{
		mbstat = *AT91C_CAN_MB6_MSR;
		if (mbstat & AT91C_CAN_MRDY)
			{
			msg.msglen = (mbstat >> 16) & 0x0F;
			msg.index  = *AT91C_CAN_MB1_MID;
			msg.mailbox= 6;
			msg.payload_lo = *AT91C_CAN_MB6_MDL;
			msg.payload_hi = *AT91C_CAN_MB6_MDH;
			*AT91C_CAN_MB6_MCR = AT91C_CAN_MTCR;
			if (xQueueSendFromISR (rx_queue, &msg, FALSE))
				{
				awakened_by_rx = TRUE;
				}
			}
		}

	//////////////////////////////////////////////////////
	//
	//				Transmit Mailbox
	//
	//////////////////////////////////////////////////////

	mbstat = *AT91C_CAN_MB7_MSR;						// Process Transmit Mailbox 7 status
	if ((mbstat & AT91C_CAN_MRDY) != 0)					//	Is output mailbox ready for more?
		{												//		Yes!
		if (xQueueReceiveFromISR (tx_queue, (void*)&msg, &awakened_by_tx))
			{
			*AT91C_CAN_MB7_MMR = AT91C_CAN_MOT_DIS;		//			Disable mailbox (needed for writing to MID)
			*AT91C_CAN_MB7_MAM = 0;						//			Enable all IDs
			*AT91C_CAN_MB7_MID = msg.index;				//			Message index
			*AT91C_CAN_MB7_MDL = msg.payload_lo;		//			Message data
			*AT91C_CAN_MB7_MDH = msg.payload_hi;
			*AT91C_CAN_MB7_MMR = AT91C_CAN_MOT_TX;		//			Re-enable it
			*AT91C_CAN_MB7_MCR = (msg.msglen << 16)		//			Set up message length
							   | AT91C_CAN_MTCR;		//			and send it
			}
		else
			{
			*AT91C_CAN_IDR	= AT91C_CAN_MB7;			// Disable interrupt until msgs are queued up
			}
		}
	else
		{
		}

	*AT91C_AIC_EOICR = 0;								// Tell chip we are done with CAN IRQs

	if (awakened_by_tx || awakened_by_rx)				// Tell OS to handle message distribution
		{
		portYIELD_FROM_ISR ();
		}
	}

//////////////////////////////////////////////////////////////////////////////
//
//							Message Queue I/O
//
//////////////////////////////////////////////////////////////////////////////
//
//	Author:		Andy Kunz
//	Date:		3-26-08
//	Modifications:	none (created for FreeRTOS 4.7.2 update)
//

static void atcan_isr_wrapper (void) __attribute__((naked));
static void atcan_isr_wrapper (void)
	{
	portSAVE_CONTEXT ();

	atcan_isr ();

	portRESTORE_CONTEXT ();
	}

//////////////////////////////////////////////////////////////////////////////
//
//							Message Queue I/O
//
//////////////////////////////////////////////////////////////////////////////
//
//	Author:		Andy Kunz
//	Date:		3-26-08
//	Modifications:	When you put a msg into the tx queue, enable the tx interrupt
//					so that the ISR can send the msg.
//
//	Author:		Andy Kunz
//	Date:		2-20-08
//	Modifications:	none
//

UINT8 atcan_put (ATCANMSG *msg)							// Queue message for transmission
	{
	UINT8	retval;
														// If there's room, insert it
	if (uxQueueMessagesWaiting (tx_queue) < CAN_TX_QUEUE_SIZE)
		{
		xQueueSend (tx_queue, msg, TX_BLOCKTIME);
		retval = ATCAN_SUCCESS;							// Tell them success
		*AT91C_CAN_IER	= AT91C_CAN_MB7;				// Enable interrupt to allow it to fly out
		}
	else												// If no room
		{
		retval = ATCAN_FAILED;							//		Tell caller it failed
		*AT91C_CAN_IER	= AT91C_CAN_MB7;				//		Make sure it wasn't because the TX isn't interrupting
		}

	return (retval);
	}

//............................................................................

UINT8 atcan_get (ATCANMSG *msg)							// Receive queued message
	{
	UINT8	retval;

	if (uxQueueMessagesWaiting (rx_queue))				// If there's a message
		{
		xQueueReceive (rx_queue, msg, RX_BLOCKTIME);	//		Grab it
		retval = ATCAN_SUCCESS;							//		Tell them success
		}
	else												// No message returns failure code
		{
		if (msg != NULL)
			memset (msg, 0, sizeof (ATCANMSG));			//		Blank out message buffer too (in case of stupid callers)
		retval = ATCAN_FAILED;
		}

	return (retval);
	}

//////////////////////////////////////////////////////////////////////////////
//
//							Buffer Status Reports
//
//////////////////////////////////////////////////////////////////////////////
//
//	Author:		Andy Kunz
//	Date:		2-20-08
//	Modifications:	none
//

UINT32 atcan_instatus (void)							// Is there anything in the receive queue?
	{
	return (uxQueueMessagesWaiting (rx_queue));			// Tell them how many messages are available
	}

//............................................................................

UINT32 atcan_outstatus (void)							// Is there room in the transmit queue?
	{
	return (CAN_TX_QUEUE_SIZE - uxQueueMessagesWaiting (tx_queue));	// Tell them how many more messages the queue can accept
	}

//////////////////////////////////////////////////////////////////////////////
//
//							Module Initialization
//
//////////////////////////////////////////////////////////////////////////////
//
//	Author:		Andy Kunz
//	Date:		3-26-08
//	Modifications:	Don't enable transmitter mailbox interrupt on startup
//
//	Author:		Andy Kunz
//	Date:		2-20-08
//	Modifications:	none
//
void atcan_init (void)									// Initialize CAN interface
	{
	rx_queue = xQueueCreate (CAN_RX_QUEUE_SIZE, sizeof (ATCANMSG));
	tx_queue = xQueueCreate (CAN_TX_QUEUE_SIZE, sizeof (ATCANMSG));

	*AT91C_CAN_MR	= 0;								// Disable CAN module

	*AT91C_PMC_PCER	= 1 << AT91C_ID_CAN;				// Enable clock to module

	*AT91C_PIOA_PDR = PIN_CANRX							// Disable PIO access to pins
					| PIN_CANTX;
	*AT91C_PIOA_ASR = PIN_CANRX							// Assign receiver pin to CAN module
					| PIN_CANTX;						// Assign transmitter pin to CAN module

	*AT91C_CAN_MR	= AT91C_CAN_CANEN;					// Enable CAN module

	*AT91C_CAN_IDR	= 0xFFFFFFFF;						// Disable all interrupt sources

	*AT91C_CAN_IER	= AT91C_CAN_MB0						// Enable interrupts from mailboxes
					| AT91C_CAN_MB1
					| AT91C_CAN_MB2
					| AT91C_CAN_MB3
					| AT91C_CAN_MB4
					| AT91C_CAN_MB5
					| AT91C_CAN_MB6;
//					| AT91C_CAN_MB7;					// Don't enable interrupts on transmitter until a message is ready to go out

#if (USE_CYGNUS_PARAMS)
	*AT91C_CAN_BR	 									// Configure baud rate
					= (CAN_BR_SMP		<< 24)
					| (CAN_BR_BRP		<< 16)
					| (CAN_BR_SJW		<< 12)
					| (CAN_BR_PROPAG	<< 8)
					| (CAN_BR_PHASE_1	<< 4)
					| (CAN_BR_PHASE_2	<< 0);
#else
	*AT91C_CAN_BR	= CAN_BR_125K;
#endif

	*AT91C_CAN_MB0_MMR	= AT91C_CAN_MOT_RX;				// Configure mailbox modes
	*AT91C_CAN_MB1_MMR	= AT91C_CAN_MOT_RX;
	*AT91C_CAN_MB2_MMR	= AT91C_CAN_MOT_RX;
	*AT91C_CAN_MB3_MMR	= AT91C_CAN_MOT_RX;
	*AT91C_CAN_MB4_MMR	= AT91C_CAN_MOT_RX;
	*AT91C_CAN_MB5_MMR	= AT91C_CAN_MOT_RX;
	*AT91C_CAN_MB6_MMR	= AT91C_CAN_MOT_RXOVERWRITE;
	*AT91C_CAN_MB7_MMR	= AT91C_CAN_MOT_TX;

	*AT91C_CAN_MB0_MAM	= 0;							// Accept all identifiers
	*AT91C_CAN_MB1_MAM	= 0;
	*AT91C_CAN_MB2_MAM	= 0;
	*AT91C_CAN_MB3_MAM	= 0;
	*AT91C_CAN_MB4_MAM	= 0;
	*AT91C_CAN_MB5_MAM	= 0;
	*AT91C_CAN_MB6_MAM	= 0;
	*AT91C_CAN_MB6_MAM	= 0;

	*AT91C_CAN_MB0_MCR = AT91C_CAN_MTCR;				//	Tell it that we're ready to receive in all mailboxes
	*AT91C_CAN_MB1_MCR = AT91C_CAN_MTCR;
	*AT91C_CAN_MB2_MCR = AT91C_CAN_MTCR;
	*AT91C_CAN_MB3_MCR = AT91C_CAN_MTCR;
	*AT91C_CAN_MB4_MCR = AT91C_CAN_MTCR;
	*AT91C_CAN_MB5_MCR = AT91C_CAN_MTCR;
	*AT91C_CAN_MB6_MCR = AT91C_CAN_MTCR;
//	*AT91C_CAN_MB7_MCR = AT91C_CAN_MTCR;

	AT91C_BASE_AIC->AIC_SVR[AT91C_ID_CAN] = (AT91_REG)atcan_isr_wrapper;	// Set vector to point to wrapper
	AT91C_BASE_AIC->AIC_SMR[AT91C_ID_CAN] = AT91C_AIC_SRCTYPE_HIGH_LEVEL	// Set mode
										  | (AT91C_AIC_PRIOR_LOWEST + 2);	// and priority
			
	*AT91C_AIC_IECR	= (1 << AT91C_ID_CAN);				// Enable this interrupt in AIC
	}

